import copy
import enum
import functools
import json
import pathlib
import threading
import typing

import calibration_client
import h5py
import numpy as np
from calibration_client.modules import (
    Calibration,
    CalibrationConstant,
    CalibrationConstantVersion,
    Detector,
    DetectorType,
    Parameter,
    PhysicalDetectorUnit,
)
from karabo.bound import DOUBLE_ELEMENT, NODE_ELEMENT, STRING_ELEMENT, UINT32_ELEMENT

from . import utils


class ConditionNotFound(Exception):
    pass


class DetectorNotFound(Exception):
    pass


class ModuleNotFound(Exception):
    pass


class CalibrationNotFound(Exception):
    pass


class CalibrationClientConfigError(Exception):
    pass


def _check_resp(resp, exception=Exception):
    # TODO: probably verify using "info" that exception is the right one
    if not resp["success"]:
        # TODO: probably more types of app_info errors?
        if resp["app_info"]:
            if "not found" in resp["info"]:
                # this was likely the exception exception
                raise exception(resp["info"])
            else:
                # but could also be authorization or similar issue
                raise CalibrationClientConfigError(resp["app_info"])
        raise exception(resp["info"])


class DetectorStandin(typing.NamedTuple):
    detector_name: str
    modno_to_source: dict
    frames_per_train: int
    module_shape: tuple


class OperatingConditions(dict):
    # TODO: support deviation?
    def encode(self):
        return {
            "parameters": [
                {
                    "parameter_name": key,
                    "lower_deviation_value": 0.0,
                    "upper_deviation_value": 0.0,
                    "flg_available": False,
                    "value": value,
                }
                for (key, value) in self.items()
            ]
        }

    def __hash__(self):
        # this takes me back to pre-screening interview time...
        return hash(tuple(sorted(self.items())))


class BaseCalcatFriend:
    _constant_enum_class = None  # subclass should set
    _constants_need_conditions = None  # subclass should set

    @staticmethod
    def add_schema(schema, detector_type, prefix="constantParameters"):
        """Add elements needed by this object to device's schema (expectedSchema)

        All elements added to schema go under prefix which should end with name of
        node which does not exist yet. To change default values and add more fields,
        extend this method in subclass.

        The values set on the device schema can handily be queried via _get.  This
        should only be done in helper methods generating "conditions"; see for example
        AgipdCalcatFriend.dark_condition.
        """

        (NODE_ELEMENT(schema).key(prefix).commit(),)

        # Parameters which any detector would probably have (extend this in subclass)
        # TODO: probably switch to floating point for everything, including mem cells
        (
            STRING_ELEMENT(schema)
            .key(f"{prefix}.detectorType")
            .displayedName("Detector type name")
            .description(
                "Name of detector type in CalCat; typically has suffix '-Type'"
            )
            .readOnly()
            .initialValue(detector_type)
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{prefix}.detectorName")
            .assignmentOptional()
            .defaultValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{prefix}.karaboDa")
            .assignmentOptional()
            .defaultValue("")
            .commit(),
            UINT32_ELEMENT(schema)
            .key(f"{prefix}.memoryCells")
            .assignmentOptional()
            .defaultValue(352)
            .reconfigurable()
            .commit(),
            UINT32_ELEMENT(schema)
            .key(f"{prefix}.pixelsX")
            .assignmentOptional()
            .defaultValue(512)
            .commit(),
            UINT32_ELEMENT(schema)
            .key(f"{prefix}.pixelsY")
            .assignmentOptional()
            .defaultValue(512)
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.biasVoltage")
            .assignmentOptional()
            .defaultValue(300)
            .reconfigurable()
            .commit(),
        )

    def __init__(
        self,
        device,
        secrets_fn: pathlib.Path,
        prefix="constantParameters",
    ):
        self.device = device
        self.prefix = prefix

        if not secrets_fn.is_file():
            self.device.log.WARN(f"Missing CalCat secrets file (expected {secrets_fn})")
        with secrets_fn.open("r") as fd:
            calcat_secrets = json.load(fd)

        self.caldb_store = pathlib.Path(calcat_secrets["caldb_store_path"])
        if not self.caldb_store.is_dir():
            raise ValueError(f"caldb_store location '{self.caldb_store}' is not dir")

        self.device.log.INFO(f"Connecting to CalCat at {calcat_secrets['base_url']}")
        base_url = calcat_secrets["base_url"]
        self.client = calibration_client.CalibrationClient(
            client_id=calcat_secrets["client_id"],
            client_secret=calcat_secrets["client_secret"],
            user_email=calcat_secrets["user_email"],
            base_api_url=f"{base_url}/api/",
            token_url=f"{base_url}/oauth/token",
            refresh_url=f"{base_url}/oauth/token",
            auth_url=f"{base_url}/oauth/authorize",
            scope="public",
            session_token=None,
        )
        self.device.log.INFO("CalCat connection established")

    def _get(self, key):
        """Helper to get value from attached device schema"""
        return self.device.get(f"{self.prefix}.{key}")

    @functools.cached_property
    def detector_id(self):
        resp = Detector.get_by_identifier(self.client, self._get("detectorName"))
        _check_resp(resp, DetectorNotFound)
        return resp["data"]["id"]

    @functools.cached_property
    def detector_type_id(self):
        resp = DetectorType.get_by_name(self.client, self._get("detectorType"))
        _check_resp(resp, DetectorNotFound)
        return resp["data"]["id"]

    # TODO: support updating mapping maybe?
    # (just means del self.pdus and the properties depending on it)
    @functools.cached_property
    def pdus(self):
        resp = PhysicalDetectorUnit.get_all_by_detector_id(
            self.client, self.detector_id
        )
        _check_resp(resp)
        for irrelevant_key in ("detector", "detector_type", "flg_available"):
            for pdu in resp["data"]:
                del pdu[irrelevant_key]
        return resp["data"]

    @functools.cached_property
    def _karabo_da_to_float_uuid(self):
        return {pdu["karabo_da"]: pdu["float_uuid"] for pdu in self.pdus}

    @functools.cached_property
    def _karabo_da_to_id(self):
        return {pdu["karabo_da"]: pdu["id"] for pdu in self.pdus}

    @utils.threadsafe_cache
    def calibration_id(self, calibration_name: str):
        resp = Calibration.get_by_name(self.client, calibration_name)
        # TODO: dump name in exception
        _check_resp(resp, CalibrationNotFound)
        return resp["data"]["id"]

    @utils.threadsafe_cache
    def condition_id(self, pdu, condition):
        # modifying condition parameter messes with cache
        condition_with_detector = copy.copy(condition)
        condition_with_detector["Detector UUID"] = pdu
        self.device.log.INFO(f"Look for condition: {condition_with_detector}")
        resp = self.client.search_possible_conditions_from_dict(
            "", condition_with_detector.encode()
        )
        _check_resp(resp, ConditionNotFound)
        return resp["data"][0]["id"]

    @utils.threadsafe_cache
    def constant_id(self, calibration_id, condition_id):
        resp = CalibrationConstant.get_by_uk(
            self.client,
            calibration_id=calibration_id,
            detector_type_id=self.detector_type_id,
            condition_id=condition_id,
        )
        _check_resp(resp)
        return resp["data"]["id"]

    def get_constant_version(self, constant, snapshot_at=None):
        # TODO: support snapshot
        # TODO: support creation time
        karabo_da = self._get("karaboDa")
        self.device.log.DEBUG(f"Going looking for {constant} for {karabo_da}")
        if karabo_da not in self._karabo_da_to_float_uuid:
            raise ModuleNotFound(f"Module map did not include {karabo_da}")
        if isinstance(constant, str):
            constant = self._constant_enum_class[constant]
        calibration_id = self.calibration_id(constant.name)
        condition = self._constants_need_conditions[constant]()
        condition_id = self.condition_id(
            self._karabo_da_to_float_uuid[karabo_da], condition
        )
        constant_id = self.constant_id(
            calibration_id=calibration_id, condition_id=condition_id
        )

        resp = CalibrationConstantVersion.get_by_uk(
            self.client,
            calibration_constant_id=constant_id,
            physical_detector_unit_id=self._karabo_da_to_id[karabo_da],
            event_at=None,
            snapshot_at=None,
        )
        _check_resp(resp)
        timestamp = resp["data"][
            "begin_at"
        ]  # TODO: check which key we like (also has begin_validity_at)
        file_path = (
            self.caldb_store / resp["data"]["path_to_file"] / resp["data"]["file_name"]
        )
        # TODO: return time stamps and such, those are useful
        # TODO: handle FileNotFoundError if we are led astray
        with h5py.File(file_path, "r") as fd:
            ary = np.array(fd[resp["data"]["data_set_name"]]["data"])
        return timestamp, ary

    def get_constant_version_and_call_me_back(
        self, constant, callback, snapshot_at=None
    ):
        """WIP. Callback function will get constant name, constant data, and hopefully
        soon also some metadata or whatever.
        """
        # TODO: thread safe caching throughout this class
        # TODO: do we want to use asyncio / "modern" async?
        def aux():
            data = self.get_constant_version(constant, snapshot_at)
            callback(constant, data)

        thread = threading.Thread(target=aux)
        thread.start()
        return thread


class AgipdConstants(enum.Enum):
    SlopesFF = enum.auto()
    ThresholdsDark = enum.auto()
    Offset = enum.auto()
    SlopesPC = enum.auto()
    BadPixelsDark = enum.auto()
    BadPixelsPC = enum.auto()
    BadPixelsFF = enum.auto()


class AgipdCalcatFriend(BaseCalcatFriend):
    _constant_enum_class = AgipdConstants

    def __init__(
        self,
        device,
        prefix="constantParameters",
    ):
        super().__init__(device, prefix)
        self._constants_need_conditions = {
            AgipdConstants.ThresholdsDark: self.dark_condition,
            AgipdConstants.Offset: self.dark_condition,
            AgipdConstants.SlopesPC: self.dark_condition,
            AgipdConstants.SlopesFF: self.illuminated_condition,
            AgipdConstants.BadPixelsDark: self.dark_condition,
            AgipdConstants.BadPixelsPC: self.dark_condition,
            AgipdConstants.BadPixelsFF: self.illuminated_condition,
        }

    @staticmethod
    def add_schema(schema, prefix="constantParameters"):
        super(AgipdCalcatFriend, AgipdCalcatFriend).add_schema(
            schema, "AGIPD-Type", prefix
        )

        (
            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.acquisitionRate")
            .assignmentOptional()
            .defaultValue(1.1)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.gainSetting")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.photonEnergy")
            .assignmentOptional()
            .defaultValue(9.2)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.gainMode")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.integrationtime")
            .assignmentOptional()
            .defaultValue(12)
            .reconfigurable()
            .commit(),
        )

    def dark_condition(self):
        res = OperatingConditions()
        res["Pixels X"] = self._get("pixelsX")
        res["Pixels Y"] = self._get("pixelsY")
        res["Memory cells"] = self._get("memoryCells")
        res["Acquisition rate"] = self._get("acquisitionRate")
        res["Sensor Bias Voltage"] = self._get("biasVoltage")
        return res

    def illuminated_condition(self):
        res = OperatingConditions()
        res["Memory cells"] = self._get("memoryCells")
        res["Sensor Bias Voltage"] = self._get("biasVoltage")
        res["Pixels X"] = self._get("pixelsX")
        res["Pixels Y"] = self._get("pixelsY")
        res["Source Energy"] = self._get("photonEnergy")
        res["Acquisition rate"] = self._get("acquisitionRate")
        res["Gain Setting"] = self._get("gainSetting")
        return res


class DsscConstants(enum.Enum):
    Offset = enum.auto()


class DsscCalcatFriend(BaseCalcatFriend):
    _constant_enum_class = DsscConstants

    def __init__(
        self,
        device,
        prefix="constantParameters",
    ):
        super().__init__(device, prefix)
        self._constants_need_conditions = {
            DsscConstants.Offset: self.dark_condition,
        }

    @staticmethod
    def add_schema(schema, prefix="constantParameters"):
        super(DsscCalcatFriend, DsscCalcatFriend).add_schema(
            schema, "DSSC-Type", prefix
        )
        (
            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.pulseIdChecksum")
            .assignmentOptional()
            .defaultValue(2.8866323107820637e-36)
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.acquisitionRate")
            .assignmentOptional()
            .defaultValue(4.5)
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.encodedGain")
            .assignmentOptional()
            .defaultValue(67328)
            .commit(),
        )

    def dark_condition(self):
        res = OperatingConditions()
        res["Memory cells"] = self._get("memoryCells")
        res["Sensor Bias Voltage"] = self._get("biasVoltage")
        res["Pixels X"] = self._get("pixelsX")
        res["Pixels Y"] = self._get("pixelsY")
        # res["Pulse id checksum"] = self._get("pulseIdChecksum")
        # res["Acquisition rate"] = self._get("acquisitionRate")
        # res["Encoded gain"] = self._get("encodedGain")
        return res
