import functools
import inspect
import threading
import time
import timeit

import numpy as np


def threadsafe_cache(fun):
    """This decorator imitates functools.cache, but threadsafer

    With multiple threads hitting a function cached by functools.cache, it is possible
    to trigger recomputation. This decorator adds granular locking: each key in the
    cache (derived from arguments) has its own lock.
    """

    locks = {}
    results = {}
    fun_sig = inspect.signature(fun)

    @functools.wraps(fun)
    def aux(*args, **kwargs):
        bound_args = fun_sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        key = bound_args.args + tuple(bound_args.kwargs.items())
        if key in results:
            return results[key]
        with locks.setdefault(key, threading.Lock()):
            if key in results:
                # someone else did this - may still be processing
                return results[key]
            else:
                res = fun(*args, **kwargs)
                results[key] = res
                return res

    return aux


_np_typechar_to_c_typestring = {
    "?": "bool",
    "B": "unsigned char",
    "D": "double complex",
    "F": "float complex",
    "G": "long double complex",
    "H": "unsigned short",
    "I": "unsigned int",
    "L": "unsigned long",
    "Q": "unsigned long long",
    "b": "char",
    "d": "double",
    "e": "half",  # warning: only in CUDA with special support
    "f": "float",
    "g": "long double",
    "h": "short",
    "i": "int",
    "l": "long",
    "q": "long long",
}


def np_dtype_to_c_type(dtype):
    as_char = np.sctype2char(dtype)
    return _np_typechar_to_c_typestring[as_char]


def enum_to_c_template(enum_class):
    res = [f"enum {enum_class.__name__} {{"]
    for field in enum_class:
        res.append(f"\t{field.name} = {field.value},")
    res.append("};")
    return "\n".join(res)


def ceil_div(num, denom):
    return (num + denom - 1) // denom


def shape_after_transpose(input_shape, transpose_pattern, squeeze=True):
    if squeeze:
        input_shape = tuple(dim for dim in input_shape if dim > 1)
    if transpose_pattern is None:
        return input_shape
    return tuple(np.array(input_shape)[list(transpose_pattern)].tolist())


class DelayableTimer:
    """Start a timer which can be extended

    Useful for reverting to state after inactivity, for instance.

    timer defaults to timeit.default_timer - it should be a timer returning
    globally increasing number of seconds.
    """

    def __init__(self, timeout, callback, timer=timeit.default_timer):
        self.timer = timer
        self.stop_time = self.timer() + timeout

        def runner():
            now = self.timer()
            while now < self.stop_time:
                diff = self.stop_time - now
                time.sleep(diff)
                now = self.timer()
            callback()

        self.thread = threading.Thread(target=runner)
        self.thread.start()

    def set_timeout(self, timeout):
        """Delay stop time to now + timeout

        If now + timeout is sooner than already set timeout, this does nothing"""
        self.stop_time = self.timer() + timeout

    def add_timeout(self, timeout):
        """Simply add timeout to current stop time"""
        self.stop_time += timeout


class RepeatingTimer:
    """Similar to DelayableTimer, but will keep running with pre-set intervals"""

    def __init__(self, interval, callback, timer=timeit.default_timer, start_now=True):
        self.timer = timer
        self.stopped = True
        self.interval = interval
        self.callback = callback
        if start_now:
            self.start()

    def delay(self):
        self.stop_time = self.timer() + self.interval

    def start(self):
        self.stopped = False
        self.stop_time = self.timer() + self.interval

        def runner():
            while not self.stopped:
                now = self.timer()
                while now < self.stop_time:
                    diff = self.stop_time - now
                    time.sleep(diff)
                    if self.stopped:
                        return
                    now = self.timer()
                self.callback()
                self.stop_time = self.timer() + self.interval

        self.thread = threading.Thread(target=runner)
        self.thread.start()

    def stop(self):
        self.stopped = True


class Throttler:
    """Similar to DelayableTimer, but will keep running with pre-set intervals"""

    def __init__(self, interval, timer=timeit.default_timer):
        self.timer = timer
        self.interval = interval
        self.latest_call = None

    def ready(self):
        if self.latest_call is None:
            return True
        else:
            return self.latest_call + self.interval <= self.timer()

    def update(self):
        self.latest_call = self.timer()

    def ready_update(self):
        time = self.timer()
        if self.latest_call is None or self.latest_call + self.interval <= time:
            self.latest_call = time
            return True
        else:
            return False
